import os

from xrdimageutil import Catalog

class TestCatalog:

    relative_path = "data/singh"
    absolute_path = os.path.abspath(path=relative_path)
    catalog_name = "test-catalog"

    def test_instantiation_with_valid_name_expects_not_none_type(self):
        catalog = Catalog(name=self.catalog_name)
        assert type(catalog) is not None

    def test_instantiation_with_invalid_name_expects_key_error(self):
        try:
            catalog = Catalog(name="invalid-name")
        except KeyError:
            assert True

    def test_get_scan_with_valid_number(self):
        catalog = Catalog(name=self.catalog_name)
        try:
            scan = catalog.get_scan(scan_id=61)
            assert True
        except:
            assert False

    def test_get_scan_with_invalid_number(self):
        catalog = Catalog(name=self.catalog_name)
        try:
            scan = catalog.get_scan(scan_id=0)
        except KeyError:
            assert True

    def test_get_scan_with_noninteger_key(self):
        catalog = Catalog(name=self.catalog_name)
        try:
            scan = catalog.get_scan(scan_id="invalid-id")
        except TypeError:
            assert True

    def test_get_scans_with_valid_list(self):
        catalog = Catalog(name=self.catalog_name)
        try:
            scans = catalog.get_scans(
                scan_ids=[61, 62, 63]
            )
            assert True
        except:
            assert False

    def test_get_scans_with_invalid_list(self):
        catalog = Catalog(name=self.catalog_name)
        try:
            scans = catalog.get_scans(
                scan_ids=[61, 62, 63, 0]
            )
            assert False
        except KeyError:
            assert True

    def test_get_scans_with_nonlist_input(self):
        catalog = Catalog(name=self.catalog_name)
        try:
            scans = catalog.get_scans(
                scan_ids=61
            )
            assert False
        except TypeError:
            assert True



"""Copyright (c) UChicago Argonne, LLC. All rights reserved.
See LICENSE file.
"""

import databroker
from prettytable import PrettyTable
import pyqtgraph as pg
import xrayutilities as xu

from xrdimageutil import utils
from xrdimageutil.gui import image_data_widget, line_data_widget


class Catalog:
    """Houses (i) a Bluesky catalog, already unpacked and (ii) a 
    dictionary of Scan objects that can be accessed.
    """
    
    bluesky_catalog = None # Bluesky dictionary-like catalog
    name = None # Local name for catalog
    scan_uid_dict = None # Dictionary of scans in catalog with UID as key
    scan_id_dict = None # Dictionary of scans in catalog with scan ID as key

    def __init__(self, name) -> None:

        self.name = str(name)
        self.bluesky_catalog = databroker.catalog[self.name]

        # Currently only configured for beamline 6-ID-B
        utils._add_catalog_handler(catalog=self)

        # Creates a Scan object for every run in the catalog
        # Adds Scans to a dictionary
        self.scan_uid_dict, self.scan_id_dict = {}, {}
        for scan_uid in sorted(list(self.bluesky_catalog)):
            scan = Scan(catalog=self, uid=scan_uid)
            self.scan_uid_dict.update({scan_uid: scan})
            try:
                self.scan_id_dict.update({scan.scan_id: scan})
            except:
                pass

        # Checks if all scan ID's are unique
        # If not, then search by scan ID is not allowed in catalog
        if len(self.scan_id_dict.keys()) != len(self.scan_uid_dict.keys()):
            self.scan_id_dict = None

    def get_scan(self, scan_id: int):
        """Returns scan from given numerical scan ID."""

        if type(scan_id) != int:
            raise TypeError(f"Scan ID {scan_id} does not exist.")

        # Checks if scan ID exists
        if scan_id not in self.scan_id_dict:
            raise KeyError(f"Scan ID {scan_id} does not exist.")

        return self.scan_id_dict[scan_id]

    def get_scans(self, scan_ids: list) -> list:
        """Returns list of scans from given numerical scan ID's."""

        if type(scan_ids) != list:
            raise TypeError("Input needs to be a list.")

        scan_list = []
        for scan_id in scan_ids:
            scan = self.get_scan(scan_id=scan_id)
            scan_list.append(scan)

        return scan_list
    
    def scan_count(self) -> int:
        """Returns the number of scans in catalog."""
        
        return len(self.scan_uid_dict.keys())

    def list_scans(self) -> None:
        """Prints formatted string table listing scans in catalog."""

        headers = [
            "scan_id", "motors", 
            "motor_start", "motor_stop", "n_pts",
            "sample", "proposal_id", "user"
        ]
        table = PrettyTable(headers)

        scan_uids = list(self.scan_uid_dict.keys())
        scans = [self.scan_uid_dict[uid] for uid in scan_uids]

        for scan in scans:
            row = [
                scan.scan_id, scan.motors, 
                scan.motor_bounds[0], scan.motor_bounds[1], scan.point_count(), 
                scan.sample, scan.proposal_id, scan.user
            ]
            table.add_row(row)

        table.sortby = "scan_id"
        print(table)

    def view_line_data(self) -> None:
        """Displays Scan line data in an interactive GUI."""
        
        self.app = pg.mkQApp()
        self.window = line_data_widget.CatalogLineDataWidget(catalog=self)
        self.window.raise_()
        self.window.show()
        self.window.raise_()
        self.app.exec_()

        
class Scan:
    """Houses data and metadata for a single scan."""

    catalog = None # Parent Catalog
    uid = None # UID for scan; given by bluesky
    bluesky_run = None # Raw Bluesky run for scan
    scan_id = None # Simple ID given to scan by user -- not always unique
    sample = None # Experimental sample
    proposal_id = None # Manually provided Proposal ID
    user = None # Experimental user
    motors = None # List of variable motors for scan
    motor_bounds = None # The starting and ending values for each variable motor
    h = None # List of H center values throughout scan
    k = None # List of K center values throughout scan
    l = None # List of L center values throughout scan
    
    rsm = None # Reciprocal space map for every point within a scan
    rsm_bounds = None # Min/max HKL values for RSM
    raw_data = None # 3D numpy array of scan data
    gridded_data = None # Interpolated and transformed scan data
    gridded_data_coords = None # HKL coordinates for gridded data

    def __init__(self, catalog: Catalog, uid: str) -> None:

        self.catalog = catalog
        self.uid = uid
        self.bluesky_run = catalog.bluesky_catalog[uid]
        self.scan_id = self.bluesky_run.metadata["start"]["scan_id"]
        self.sample = self.bluesky_run.metadata["start"]["sample"]
        self.proposal_id = self.bluesky_run.metadata["start"]["proposal_id"]
        self.user = self.bluesky_run.metadata["start"]["user"]
        self.motors = self.bluesky_run.metadata["start"]["motors"]
        self.motor_bounds = utils._get_motor_bounds(self)
        self.h, self.k, self.l = utils._get_hkl_centers(self)

        # TODO: Figure out where to put these steps
        self.rsm = utils._get_rsm_for_scan(self)
        self.rsm_bounds = utils._get_rsm_bounds(self)
        self.raw_data = utils._get_raw_data(self)

    def point_count(self) -> int:
        """Returns number of points in scan."""

        if "primary" not in self.bluesky_run.keys():
            return 0
        else:
            return self.bluesky_run.primary.metadata["dims"]["time"]

    def grid_data(
        self,
        h_count: int=250, k_count: int=250, l_count: int=250,
        h_min: float=None, h_max: float=None, 
        k_min: float=None, k_max: float=None,
        l_min: float=None, l_max: float=None
    ) -> None:
        """Constructs gridded 3D image from RSM coordinates."""

        # Provided bounds for gridding
        grid_bounds = [h_min, h_max, k_min, k_max, l_min, l_max]

        # Bounds in reciprocal space map, reshaped to a list
        rsm_bounds = [self.rsm_bounds[b] for b in list(self.rsm_bounds.keys())]
        
        for i in range(len(grid_bounds)):
            if grid_bounds[i] is None:
                grid_bounds[i] = rsm_bounds[i]

        h_map = self.rsm[:, :, :, 0]
        k_map = self.rsm[:, :, :, 1]
        l_map = self.rsm[:, :, :, 2]

        # Prepares gridder bounds/interpolation
        gridder = xu.Gridder3D(
            nx=h_count, 
            ny=k_count, 
            nz=l_count
        )
        gridder.KeepData(True)
        gridder.dataRange(
            xmin=grid_bounds[0], xmax=grid_bounds[1],
            ymin=grid_bounds[2], ymax=grid_bounds[3],
            zmin=grid_bounds[4], zmax=grid_bounds[5],
            fixed=True
        )

        # Grids raw data with bounds
        gridder(h_map, k_map, l_map, self.raw_data)
        self.gridded_data = gridder.data

        # Retrieves HKL coordinates for gridded data
        self.gridded_data_coords = [gridder.xaxis, gridder.yaxis, gridder.zaxis]

    def view_image_data(self) -> None:
        """Displays Scan image data in an interactive GUI."""
        
        self.app = pg.mkQApp()
        self.window = image_data_widget.ScanImageDataWidget(scan=self)
        self.window.raise_()
        self.window.show()
        self.window.raise_()
        self.app.exec_()



import os

from xrdimageutil import Catalog, Scan

class TestScan:

    relative_path = "data/singh"
    absolute_path = os.path.abspath(path=relative_path)
    catalog_name = "test-catalog"
        
    def test_point_count_expects_30(self):
        catalog = Catalog(name=self.catalog_name)
        scan = catalog.get_scan(61)

        if scan.point_count() == 30:
            assert True



"""Copyright (c) UChicago Argonne, LLC. All rights reserved.
See LICENSE file.
"""

import area_detector_handlers.handlers as adh
from dask.array import from_array
import numpy as np
import pyqtgraph as pg
from sklearn import preprocessing
import xrayutilities as xu


# TODO: Document everything

class PilatusHDF5Handler(adh.AreaDetectorHDF5SingleHandler):
    """Handler for the Pilatus detector HDF5 files. This version is
    geared specifically towards beamline 6-ID-B.

    Originally from Gilberto Fabbris.
    """

    specs = (
        {"AD_HDF5_Pilatus_6idb"} | adh.AreaDetectorHDF5SingleHandler.specs
    )

    def __call__(self, point_number):
        return from_array(super().__call__(point_number))


def _add_catalog_handler(catalog) -> None:
    bluesky_catalog = catalog.bluesky_catalog
    bluesky_catalog.register_handler(
        "AD_HDF5_Pilatus_6idb", 
        PilatusHDF5Handler, 
        overwrite=True
    )


def _get_rsm_for_scan(scan):
    """Returns a reciprocal space map for a given scan.
    
    Currently restricted to four circle geometries
    """

    run = scan.bluesky_run

    # Checks if scan includes a "primary" category,
    # where data is traditionally stored
    if "primary" not in run.keys():
        return None

    # Instrument config values
    omega_values = run.primary.read()["fourc_omega"].values
    chi_values = run.primary.read()["fourc_chi"].values
    phi_values = run.primary.read()["fourc_phi"].values
    tth_values = run.primary.read()["fourc_tth"].values
    sample_circle_values = [omega_values, chi_values, phi_values]
    instrument_circle_values = [tth_values]
    circle_values = sample_circle_values + instrument_circle_values
    energy_values = run.primary.config["fourc"].read()["fourc_energy"].values * 1000
    ub_matrix = run.primary.config["fourc"].read()["fourc_UB"].values[0]
    sample_circle_directions = ["z-", "y+", "z-"]
    detector_circle_directions = ["z-"]
    primary_beam_direction = [0, 1, 0]
    inplane_reference_direction = [0, 1, 0]
    sample_normal_direction = [0, 0, 1]

    # Detector config values
    pixel_directions = ["z-", "x-"]
    center_channel_pixels = [252, 107]
    pixel_count = [
        run.primary.metadata["dims"]["dim_2"], 
        run.primary.metadata["dims"]["dim_1"]
    ]
    detector_size = [83.764, 33.54]
    pixel_size = [
        detector_size[0] / pixel_count[0],
        detector_size[1] / pixel_count[1]
    ]
    detector_distance = 900.644 # mm
    roi = [0, pixel_count[0], 0, pixel_count[1]]

    q_conversion = xu.experiment.QConversion(
        sampleAxis=sample_circle_directions,
        detectorAxis=detector_circle_directions,
        r_i=primary_beam_direction
    )

    point_rsm_list = []
    for i in range(scan.point_count()):
        
        hxrd = xu.HXRD(
            idir=inplane_reference_direction,
            ndir=sample_normal_direction,
            en=energy_values[i],
            qconv=q_conversion
        )

        hxrd.Ang2Q.init_area(
            pixel_directions[0], pixel_directions[1],
            cch1=center_channel_pixels[0], cch2=center_channel_pixels[1],
            Nch1=pixel_count[0], Nch2=pixel_count[1],
            pwidth1=pixel_size[0], pwidth2=pixel_size[1],
            distance=detector_distance, roi=roi
        )

        point_circle_values = [circle[i] for circle in circle_values]
        qx, qy, qz = hxrd.Ang2Q.area(*point_circle_values, UB=ub_matrix)

        point_rsm = np.array([qx, qy, qz])
        point_rsm_list.append(point_rsm)
    
    # Converts to ndarray and reorders dimensions
    rsm = np.array(point_rsm_list)
    rsm = rsm.swapaxes(1, 3)
    rsm = rsm.swapaxes(1, 2)

    return rsm


def _get_rsm_bounds(scan):
    """Returns the minimum and maximum values for H, K, and L with an RSM."""

    rsm = scan.rsm
    rsm_bounds = {}

    if scan.rsm is None:
        rsm_bounds.update({"h_min": None})
        rsm_bounds.update({"h_max": None})
        rsm_bounds.update({"k_min": None})
        rsm_bounds.update({"k_max": None})
        rsm_bounds.update({"l_min": None})
        rsm_bounds.update({"l_max": None})
    else:
        rsm_bounds.update({"h_min": np.amin(rsm[:, :, :, 0])})
        rsm_bounds.update({"h_max": np.amax(rsm[:, :, :, 0])})
        rsm_bounds.update({"k_min": np.amin(rsm[:, :, :, 1])})
        rsm_bounds.update({"k_max": np.amax(rsm[:, :, :, 1])})
        rsm_bounds.update({"l_min": np.amin(rsm[:, :, :, 2])})
        rsm_bounds.update({"l_max": np.amax(rsm[:, :, :, 2])})
        
    return rsm_bounds


def _get_motor_bounds(scan) -> tuple:
    """Returns minimum and maximum values for each instrument motor."""

    run = scan.bluesky_run

    if "primary" not in run.keys():
        return (None, None)

    motors = scan.motors
    m_start = []
    m_stop = []
    start, stop = 0, 0
    for motor in motors:
        start = run.primary.read()[motor].values[0]
        stop = run.primary.read()[motor].values[-1]

        m_start.append(round(start, 5))
        m_stop.append(round(stop, 5))

    return (m_start, m_stop)


def _get_raw_data(scan) -> np.ndarray:
    """Returns raw detector images for a Scan."""
    run = scan.bluesky_run

    if "primary" not in run.keys():
        return None
    
    raw_data_unordered = run.primary.read()["pilatus100k_image"].values
    raw_data_unordered = np.squeeze(raw_data_unordered)
    raw_data = np.swapaxes(raw_data_unordered, 1, 2)

    return raw_data


def _get_hkl_centers(scan) -> tuple:
    """Returns lists of H, K, and L center values."""

    run = scan.bluesky_run

    if "primary" not in run.keys():
        return (None, None, None)
    
    h = run.primary.read()["fourc_h"].values
    k = run.primary.read()["fourc_k"].values
    l = run.primary.read()["fourc_l"].values

    return (h, k, l)


def _create_colormap(
    name: str,
    scale: str,
    min: float=0.0,
    max: float=1.0,
    n_pts: int=16,
    base: float=1.75,
    gamma: float=2
) -> pg.ColorMap:
    """Returns a color map object created from given parameters."""

    if name in pg.colormap.listMaps(source="matplotlib"):
        colors = pg.colormap.getFromMatplotlib(name).getLookupTable(nPts=n_pts)
    elif name in pg.colormap.listMaps(source="colorcet"):
        colors = pg.colormap.getFromColorcet(name).getLookupTable(nPts=n_pts)
    elif name in pg.colormap.listMaps():
        colors = pg.get(name).getLookupTable(nPts=n_pts)
    else:
        raise KeyError("Color map not found.")

    if scale == "linear":
        stops = np.linspace(start=min, stop=max, num=n_pts)
        stops = np.array([list(stops)])
        stops = preprocessing.normalize(stops, norm="max")
        stops = list(stops[0])
    elif scale == "log":
        stops = np.logspace(
            start=0,
            stop=7.5,
            endpoint=True,
            num=n_pts,
            base=base
        )
        stops = np.array([list(stops)])
        stops = preprocessing.normalize(stops, norm="max")
        stops = list(stops[0])
    elif scale == "power":
        stops = np.linspace(start=min, stop=max, num=n_pts)
        stops -= min
        stops[stops < 0] = 0
        np.power(stops, gamma, stops)
        stops /= (max - min) ** gamma
        stops = np.array([list(stops)])
        stops = preprocessing.normalize(stops, norm="max")
        stops = list(stops[0])
    else:
        raise ValueError("Scale type not valid.")

    return pg.ColorMap(pos=stops, color=colors)
